(* version 1.2.1 (c) Copyright 2006 Pierre Corbineau, Jean-Baptiste Guillaumin *)
(* published under the GNU General Public License *)

open Sys;;
open Unix;;
open Printf;;

module type Entity =
sig

type t

val poids : t -> int

val dist : t -> t -> int

end

let min a b c = 
 ( let min_ab = if a < b then a else b in
    if c < min_ab then c else min_ab :int)

module Word(E:Entity):Entity with type t = E.t array =
struct

  type t = E.t array

  let dist w1 w2 =
    let hauteur = Array.length w1 in
    let largeur = Array.length w2 in
    let poids1 = Array.map E.poids w1 in
    let poids2 = Array.map E.poids w2 in
    let rangee = Array.create (1+largeur) 0 in
    let diagonale = ref 0 in  
      for i = 0 to largeur-1 do
	let result = rangee.(i) + poids2.(i) in
	  rangee.(i+1)<- result;
      done;
      for j = 0 to hauteur-1 do
	diagonale:= rangee.(0);
	let result = rangee.(0) + poids1.(j) in
	  rangee.(0)<- result;
	  for i = 0 to largeur-1 do
	    let result = min 
            (rangee.(i) + poids2.(i)) 
	    (rangee.(i+1) + poids1.(j))
	    (!diagonale + E.dist w1.(j) w2.(i)) in
	      diagonale:=rangee.(i+1);
	      rangee.(i+1)<- result;
	  done;
      done;
      rangee.(largeur)
	
  let poids w = dist w [||] 

end

module Lettre =
struct
  type t = char     
  let poids c = 2
  let dist c1 c2 =
    if c1 = c2 then 0 else
      if c1 = 'v' && c2 = 'u' then 0 else
      if c1 = 'u' && c2 = 'v' then 0 else
      if c1 = 'i' && c2 = 'j' then 0 else
      if c1 = 'j' && c2 = 'i' then 0 else
      if c1 = 'e' && c2 = '' then 1 else
      if c1 = '' && c2 = 'e' then 1 else
      if c1 = '*' || c2 = '*' then 1 else 2
    
end

module Mot = Word(Lettre)

module Paragraphe =  Word(Mot)

module Texte = Word(Paragraphe)

let is_letter = function
    'a'..'z' | 'A'..'Z' | '*' | '_' | ''  -> true
  | _ -> false

let read_words file =
  let chan = open_in file in
  let word_stack = Stack.create () in
  let buf = Buffer.create 64 in
  try 
    while true do
      let c = input_char chan in
	if is_letter c then
	  Buffer.add_char buf (Char.lowercase c)
	else
	  begin
	    if Buffer.length buf > 0 then
	      Stack.push (Buffer.contents buf) word_stack;
	    Buffer.clear buf
	  end
    done;
    assert false
  with End_of_file ->
    let _ = close_in chan in
    let n_words = Stack.length word_stack in
    let text = Array.create n_words [||] in
      for i=(n_words-1) downto 0 do
	let w = Stack.pop word_stack in
	let wt=Array.init (String.length w) (fun n -> w.[n]) in
	  text.(i)<- wt
      done;
      assert (Stack.length word_stack = 0);
      text

let calcul file1 file2 = 
  let text1 = read_words file1 in
  let text2 = read_words file2 in
  let dist = Paragraphe.dist text1 text2 in
  let resultat = (float_of_int(dist)/.2.0) in
    resultat
   
let main () =
  let nombre_fichiers = ref ((Array.length Sys.argv) - 1) in
  let matricedist = Array.make_matrix (!nombre_fichiers + 1) (!nombre_fichiers + 1) 0. in

  for j = 1 to !nombre_fichiers do
    for i = j+1 to !nombre_fichiers do
      let res = calcul Sys.argv.(i) Sys.argv.(j) in
        matricedist.(i).(j) <- res;
        matricedist.(j).(i) <- res;
    done; 
  done;

(*on attribue le n d'argv  chaque ligne/colonne pour pouvoir s'y retrouver ensuite*)

for i = 0 to !nombre_fichiers do
  matricedist.(0).(i) <- float_of_int(i);
  matricedist.(i).(0) <- float_of_int(i);
done;

(*On copie Sys.argv dans elements, pour pouvoir travailler ensuite sur ce tableau*)

let elements = ref Sys.argv in

(*On garde en mmoire la moyenne de chaque paire d'OTU traite, afin de la sousrtaire au moment o on calcule la distance entre deux noeuds. On l'initialise  zro (ainsi les moyennes d'OTU non encore traites ne sont pas prises en compte).*)

let moyenne = Array.create (!nombre_fichiers + 1) 0. in

(*on calcule la somme des longueurs des branches, totalbranches*)

let totalbranches = Array.make_matrix (!nombre_fichiers +1) (!nombre_fichiers + 1) 0. in

let choix = ref [|0;0|] in

for etapes = !nombre_fichiers downto 3 do

(*ligne suivante pas terrible...*)

let minimum = ref (100000.) in

for j = 1 to !nombre_fichiers do
  for i = 1 to j-1 do

    let somme1 = ref (0.) in
    let somme2 = ref (0.) in

      for k = 1 to !nombre_fichiers do
        if k != i then
          somme1 := !somme1 +. matricedist.(i).(k);
        if k!= j then
          somme2 := !somme2 +. matricedist.(j).(k);
      done;

    let somme = (float_of_int(!nombre_fichiers)-.2.)*.matricedist.(i).(j) -. !somme1 -. !somme2 in

          totalbranches.(i).(j) <- somme;

      if somme < !minimum then
        begin
        minimum := somme;
        choix := [|i;j|];
        end
  done;
done;

(*Calcul de la distance par rapport au noeud*)

    let somme3 = ref (0.) in
    let somme4 = ref (0.) in

for k = 1 to !nombre_fichiers do
  if k != !choix.(0) && k != !choix.(1) then 
    begin
      somme3 := !somme3 +. matricedist.(!choix.(0)).(k);
      somme4 := !somme4 +. matricedist.(!choix.(1)).(k);
    end 
done;

let distance1 = ((matricedist.(!choix.(0)).(!choix.(1))) +. (!somme3/.(float_of_int(!nombre_fichiers) -.2.)) -. (!somme4/.(float_of_int(!nombre_fichiers) -.2.)))/.2. in

(*let distance2 = ((matricedist.(!choix.(0)).(!choix.(1))) +. (!somme4/.(float_of_int(!nombre_fichiers) -.2.)) -. (!somme3/.(float_of_int(!nombre_fichiers) -.2.)))/.2. in*)

let distance2 = matricedist.(!choix.(0)).(!choix.(1)) -. distance1 in

(*On calcule la longueur entre deux noeuds en retirant la moiti de la distance entre les OTUs qui leur sont relis (cette demi-distance a t pralablement calcule, pour les cas o elle existe, dans le tableau moyenne).*)

let anciennemoyenne1 = moyenne.(!choix.(0)) in
let anciennemoyenne2 = moyenne.(!choix.(1)) in

let distancenoeuds1 = distance1-.anciennemoyenne1 in
let distancenoeuds2 = distance2-.anciennemoyenne2 in

(*On redfinit le tableau moyenne en supprimant la colonne choix.(1) et en dcalant les suivantes (de mme que dans l'tape "reduction step" prsente ci-dessous pour la matrice des distances).*)

moyenne.(!choix.(0)) <- (matricedist.(!choix.(0)).(!choix.(1)))/.2.;
for c = !choix.(1) to !nombre_fichiers-1 do
  moyenne.(c) <- moyenne.(c+1);
done;

(*On enregistre le noeud et les distances calcules dans la colonne du tableau elements qui correspond au premier choix (!choix.(0)).*)

if !nombre_fichiers == 3 then 
!elements.(int_of_float(matricedist.(!choix.(0)).(0))) <- "(" ^ !elements.(int_of_float(matricedist.(!choix.(0)).(0))) ^ ":" ^ string_of_float(distancenoeuds1) ^ "," ^ !elements.(int_of_float(matricedist.(!choix.(1)).(0))) ^ ":" ^ string_of_float(distancenoeuds2)

else
!elements.(int_of_float(matricedist.(!choix.(0)).(0))) <- "(" ^ !elements.(int_of_float(matricedist.(!choix.(0)).(0))) ^ ":" ^ string_of_float(distancenoeuds1) ^ "," ^ !elements.(int_of_float(matricedist.(!choix.(1)).(0))) ^ ":" ^ string_of_float(distancenoeuds2) ^ ")";


(*Reduction step: on place la paire choisie  la place du premier choix (!choix.(O)). On supprime le deuxime choix (!choix.(1)) et on dcale les colonnes suivantes de la matrice.*)

nombre_fichiers := !nombre_fichiers -1;

let matricereduction = Array.make_matrix (!nombre_fichiers + 1) (!nombre_fichiers + 1) 0. in

for i = 0 to !nombre_fichiers do
  for j = 0 to i-1 do
    if i < !choix.(0) then 
      begin
        matricereduction.(i).(j) <- matricedist.(i).(j);
      end 
    else if i == !choix.(0) then 
      begin
        if j == 0 then matricereduction.(i).(j) <- matricedist.(i).(j) else
        matricereduction.(i).(j) <- (matricedist.(j).(!choix.(0)) +. matricedist.(j).(!choix.(1)))/.2.;
      end 
    else if i > !choix.(0) && i < !choix.(1) then 
      begin
        if j == !choix.(0) then matricereduction.(i).(j) <- (matricedist.(i).(!choix.(0)) +. matricedist.(i).(!choix.(1)))/.2. else
        matricereduction.(i).(j) <- matricedist.(i).(j);
      end
    else if i >= !choix.(1) then 
      begin
        if j == !choix.(0) then matricereduction.(i).(j) <- (matricedist.(i+1).(!choix.(0)) +. matricedist.(i+1).(!choix.(1)))/.2.
        else if j < !choix.(1) then
        matricereduction.(i).(j) <- matricedist.(i+1).(j)
        else if j >= !choix.(1) then
            matricereduction.(i).(j) <- matricedist.(i+1).(j+1)
      end;
      matricereduction.(j).(i) <- matricereduction.(i).(j);
 done;
 done;


(*On copie matricereduction dans matricedist et on relance l'algorithme*)

for i = 0 to !nombre_fichiers do
  for j = 0 to i-1 do
    matricedist.(i).(j) <- matricereduction.(i).(j);
    matricedist.(j).(i) <- matricedist.(i).(j);
  done;
done;

 done;

(*On ajoute ici la valeur qui reste  la fin de la boucle*)

let distancenoeuds = matricedist.(1).(2)-.moyenne.(1)-.moyenne.(2) in


!elements.(!choix.(0)) <- !elements.(int_of_float(matricedist.(!choix.(0)).(0))) ^ "," ^ (!elements.(int_of_float(matricedist.(0).(2)))) ^ ":" ^ (string_of_float(distancenoeuds)) ^ ");\n";

printf "%s" (!elements.(!choix.(0)));

    exit 0;;
main ();;

